/******************************************************************************

$Author$
  
$Modtime$
$Revision$

Description: Interface of class "CBitArray"
             (MFC template class CArray "specialized" for bit)

             The implementation of the private member functions "Copylr",
             "Copyrl", and "X" is derived from "The Standard C++ Library",
             copyright  1995 by P.J. Plauger, published by Prentice-Hall, and
             are used with permission.

$Log$

******************************************************************************/

/*** Declaration of class "CBitArray" ****************************************/
class CBitArray: public CObject  
{
  public:
  typedef ULONG T;
  enum {T_BIT = CHAR_BIT * sizeof(T), DefaultGrowByT = 2};

	class reference
  {
	  friend class CBitArray;

	  public:
		reference& operator=(bool bNewElement)                  // for b[i] = x;
    {
      m_pBitArray->SetAt(m_nIndex, bNewElement);
			return *this;
    }

		reference& operator=(const reference& r)                // for b[i] = b[j];
		{
      m_pBitArray->SetAt(m_nIndex, r);
			return *this;
    }

		operator bool() const {return m_pBitArray->GetAt(m_nIndex);}
                                                            // for x = b[i];

	  private:
		reference(CBitArray& bitArray , int nIndex): m_pBitArray(&bitArray),
                                                 m_nIndex(nIndex)
    {}

		CBitArray* m_pBitArray;
    int        m_nIndex;
	};

  CBitArray(): m_nSize(0), m_nTAllocated(0), m_pData(0) {}

  CBitArray(const CBitArray& src): m_nSize(0), m_nTAllocated(0), m_pData(0)
  {
    Copy(src);
  }

  virtual ~CBitArray() {free(m_pData);}

  int  Add      (bool bNewElement);
  int  Append   (const CBitArray& src);
  void Copy     (const CBitArray& src);
  void FreeExtra();

  bool GetAt(int nIndex) const
  {
    ASSERT(nIndex >= 0 && nIndex < m_nSize);

    return (m_pData[nIndex / T_BIT] &
            static_cast<T>(1) << (T_BIT - 1 - nIndex % T_BIT)) != 0;
  }

  const T*   GetData      () const {return m_pData;}
        T*   GetData      ()       {return m_pData;}
        int  GetSize      () const {return m_nSize;}
        int  GetUpperBound() const {return m_nSize - 1;}
        void InsertAt     (int nIndex, bool bNewElement, int nCount = 1);
        void InsertAt     (int nStartIndex, CBitArray* pNewArray);
        void RemoveAt     (int nIndex, int nCount = 1);
        void RemoveAll    ();
        void SetAt        (int nIndex, bool bNewElement);
        void SetAtGrow    (int nIndex, bool bNewElement);
        void SetSize      (int nNewSize, int nGrowBy = -1);

	reference operator[](int nIndex) {return reference(*this, nIndex);}

  private:
  void Copylr(const CBitArray& a, int nDstIndex, int nSrcIndex, int n);
  void Copyrl(const CBitArray& a, int nDstIndex, int nSrcIndex, int n);

  T X(int nIndex, int nBitPos) const
  {
    ASSERT(nIndex >= 0 && m_nSize > 0);

    return nBitPos == 0 ? m_pData[nIndex] : (m_nSize - 1) / T_BIT == nIndex ?
           m_pData[nIndex] << nBitPos :
           m_pData[nIndex] << nBitPos |
           m_pData[nIndex + 1] >> (T_BIT - nBitPos);
  }

  int m_nSize;
  int m_nTAllocated;
  T*  m_pData;
};
